#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from heat.common import exception
from heat.common.i18n import _
from heat.engine.cfn import functions as cfn_funcs
from heat.engine.cfn import parameters as cfn_params
from heat.engine import function
from heat.engine import parameters
from heat.engine import rsrc_defn
from heat.engine import template_common


class CfnTemplateBase(template_common.CommonTemplate):
    """The base implementation of cfn template."""

    SECTIONS = (
        VERSION, ALTERNATE_VERSION,
        DESCRIPTION, MAPPINGS, PARAMETERS, RESOURCES, OUTPUTS,
    ) = (
        'AWSTemplateFormatVersion', 'HeatTemplateFormatVersion',
        'Description', 'Mappings', 'Parameters', 'Resources', 'Outputs',
    )

    OUTPUT_KEYS = (
        OUTPUT_DESCRIPTION, OUTPUT_VALUE,
    ) = (
        'Description', 'Value',
    )

    SECTIONS_NO_DIRECT_ACCESS = set([PARAMETERS, VERSION, ALTERNATE_VERSION])

    _RESOURCE_KEYS = (
        RES_TYPE, RES_PROPERTIES, RES_METADATA, RES_DEPENDS_ON,
        RES_DELETION_POLICY, RES_UPDATE_POLICY, RES_DESCRIPTION,
    ) = (
        'Type', 'Properties', 'Metadata', 'DependsOn',
        'DeletionPolicy', 'UpdatePolicy', 'Description',
    )

    functions = {
        'Fn::FindInMap': cfn_funcs.FindInMap,
        'Fn::GetAZs': cfn_funcs.GetAZs,
        'Ref': cfn_funcs.Ref,
        'Fn::GetAtt': cfn_funcs.GetAtt,
        'Fn::Select': cfn_funcs.Select,
        'Fn::Join': cfn_funcs.Join,
        'Fn::Base64': cfn_funcs.Base64,
    }

    deletion_policies = {
        'Delete': rsrc_defn.ResourceDefinition.DELETE,
        'Retain': rsrc_defn.ResourceDefinition.RETAIN,
        'Snapshot': rsrc_defn.ResourceDefinition.SNAPSHOT
    }

    HOT_TO_CFN_RES_ATTRS = {'type': RES_TYPE,
                            'properties': RES_PROPERTIES,
                            'metadata': RES_METADATA,
                            'depends_on': RES_DEPENDS_ON,
                            'deletion_policy': RES_DELETION_POLICY,
                            'update_policy': RES_UPDATE_POLICY}

    HOT_TO_CFN_OUTPUT_ATTRS = {'description': OUTPUT_DESCRIPTION,
                               'value': OUTPUT_VALUE}

    def __getitem__(self, section):
        """Get the relevant section in the template."""
        if section not in self.SECTIONS:
            raise KeyError(_('"%s" is not a valid template section') % section)
        if section in self.SECTIONS_NO_DIRECT_ACCESS:
            raise KeyError(
                _('Section %s can not be accessed directly.') % section)

        if section == self.DESCRIPTION:
            default = 'No description'
        else:
            default = {}

        # if a section is None (empty yaml section) return {}
        # to be consistent with an empty json section.
        return self.t.get(section) or default

    def param_schemata(self, param_defaults=None):
        params = self.t.get(self.PARAMETERS) or {}
        pdefaults = param_defaults or {}
        for name, schema in params.items():
            if name in pdefaults:
                params[name][parameters.DEFAULT] = pdefaults[name]

        return dict((name, parameters.Schema.from_dict(name, schema))
                    for name, schema in params.items())

    def get_section_name(self, section):
        return section

    def parameters(self, stack_identifier, user_params, param_defaults=None):
        return cfn_params.CfnParameters(stack_identifier, self,
                                        user_params=user_params,
                                        param_defaults=param_defaults)

    def resource_definitions(self, stack):
        resources = self.t.get(self.RESOURCES) or {}

        conditions = self.conditions(stack)

        def defns():
            for name, snippet in resources.items():
                try:
                    defn_data = dict(self._rsrc_defn_args(stack, name,
                                                          snippet))
                except (TypeError, ValueError, KeyError) as ex:
                    msg = str(ex)
                    raise exception.StackValidationFailed(message=msg)

                defn = rsrc_defn.ResourceDefinition(name, **defn_data)
                cond_name = defn.condition()

                if cond_name is not None:
                    try:
                        enabled = conditions.is_enabled(cond_name)
                    except ValueError as exc:
                        path = [self.RESOURCES, name, self.RES_CONDITION]
                        message = str(exc)
                        raise exception.StackValidationFailed(path=path,
                                                              message=message)
                    if not enabled:
                        continue

                yield name, defn

        return dict(defns())

    def add_resource(self, definition, name=None):
        if name is None:
            name = definition.name
        hot_tmpl = definition.render_hot()

        if self.t.get(self.RESOURCES) is None:
            self.t[self.RESOURCES] = {}

        cfn_tmpl = dict((self.HOT_TO_CFN_RES_ATTRS[k], v)
                        for k, v in hot_tmpl.items())

        dep_list = cfn_tmpl.get(self.RES_DEPENDS_ON, [])
        if len(dep_list) == 1:
            dep_res = cfn_tmpl[self.RES_DEPENDS_ON][0]
            if dep_res in self.t[self.RESOURCES]:
                cfn_tmpl[self.RES_DEPENDS_ON] = dep_res
            else:
                del cfn_tmpl[self.RES_DEPENDS_ON]
        elif dep_list:
            cfn_tmpl[self.RES_DEPENDS_ON] = [d for d in dep_list
                                             if d in self.t[self.RESOURCES]]

        self.t[self.RESOURCES][name] = cfn_tmpl

    def add_output(self, definition):
        hot_op = definition.render_hot()
        cfn_op = dict((self.HOT_TO_CFN_OUTPUT_ATTRS[k], v)
                      for k, v in hot_op.items())

        if self.t.get(self.OUTPUTS) is None:
            self.t[self.OUTPUTS] = {}
        self.t[self.OUTPUTS][definition.name] = cfn_op


class CfnTemplate(CfnTemplateBase):

    CONDITIONS = 'Conditions'
    SECTIONS = CfnTemplateBase.SECTIONS + (CONDITIONS,)
    SECTIONS_NO_DIRECT_ACCESS = (CfnTemplateBase.SECTIONS_NO_DIRECT_ACCESS |
                                 set([CONDITIONS]))

    RES_CONDITION = 'Condition'
    _RESOURCE_KEYS = CfnTemplateBase._RESOURCE_KEYS + (RES_CONDITION,)
    HOT_TO_CFN_RES_ATTRS = CfnTemplateBase.HOT_TO_CFN_RES_ATTRS
    HOT_TO_CFN_RES_ATTRS.update({'condition': RES_CONDITION})

    OUTPUT_CONDITION = 'Condition'
    OUTPUT_KEYS = CfnTemplateBase.OUTPUT_KEYS + (OUTPUT_CONDITION,)

    functions = {
        'Fn::FindInMap': cfn_funcs.FindInMap,
        'Fn::GetAZs': cfn_funcs.GetAZs,
        'Ref': cfn_funcs.Ref,
        'Fn::GetAtt': cfn_funcs.GetAtt,
        'Fn::Select': cfn_funcs.Select,
        'Fn::Join': cfn_funcs.Join,
        'Fn::Split': cfn_funcs.Split,
        'Fn::Replace': cfn_funcs.Replace,
        'Fn::Base64': cfn_funcs.Base64,
        'Fn::MemberListToMap': cfn_funcs.MemberListToMap,
        'Fn::ResourceFacade': cfn_funcs.ResourceFacade,
        'Fn::If': cfn_funcs.If,
    }

    condition_functions = {
        'Fn::Equals': cfn_funcs.Equals,
        'Ref': cfn_funcs.ParamRef,
        'Fn::FindInMap': cfn_funcs.FindInMap,
        'Fn::Not': cfn_funcs.Not,
        'Fn::And': cfn_funcs.And,
        'Fn::Or': cfn_funcs.Or
    }

    def __init__(self, tmpl, template_id=None, files=None, env=None):
        super(CfnTemplate, self).__init__(tmpl, template_id, files, env)

        self.merge_sections = [self.PARAMETERS, self.CONDITIONS]

    def _get_condition_definitions(self):
        return self.t.get(self.CONDITIONS, {})

    def _rsrc_defn_args(self, stack, name, data):
        for arg in super(CfnTemplate, self)._rsrc_defn_args(stack, name, data):
            yield arg

        parse_cond = functools.partial(self.parse_condition, stack)

        yield ('condition',
               self._parse_resource_field(self.RES_CONDITION,
                                          (str, bool,
                                           function.Function),
                                          'string or boolean',
                                          name, data, parse_cond))


class HeatTemplate(CfnTemplateBase):
    functions = {
        'Fn::FindInMap': cfn_funcs.FindInMap,
        'Fn::GetAZs': cfn_funcs.GetAZs,
        'Ref': cfn_funcs.Ref,
        'Fn::GetAtt': cfn_funcs.GetAtt,
        'Fn::Select': cfn_funcs.Select,
        'Fn::Join': cfn_funcs.Join,
        'Fn::Split': cfn_funcs.Split,
        'Fn::Replace': cfn_funcs.Replace,
        'Fn::Base64': cfn_funcs.Base64,
        'Fn::MemberListToMap': cfn_funcs.MemberListToMap,
        'Fn::ResourceFacade': cfn_funcs.ResourceFacade,
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat_integrationtests.functional import functional_base

template_subnet_old_network = """
heat_template_version: 2016-10-14
parameters:
  net_cidr:
    type: string
resources:
  net:
    type: OS::Neutron::Net
  subnet:
    type: OS::Neutron::Subnet
    properties:
      cidr: { get_param: net_cidr }
      network_id: { get_resource: net }
"""

template_with_get_attr = """
heat_template_version: 2016-10-14
description: Test template to create/update subnet with translation
parameters:
  net_cidr:
    type: string
resources:
  net:
    type: OS::Neutron::Net
  net_value:
    type: OS::Heat::Value
    properties:
      value: { get_resource: net }
  subnet:
    type: OS::Neutron::Subnet
    properties:
      network: { get_attr: [net_value, value] }
      cidr: { get_param: net_cidr }
"""

template_value_from_nested_stack_main = """
heat_template_version: 2016-10-14
parameters:
  flavor:
    type: string
  image:
    type: string
  public_net:
    type: string
resources:
  network_settings:
    type: network.yaml
    properties:
      public_net: { get_param: public_net }
  server:
    type: OS::Nova::Server
    properties:
      flavor: { get_param: flavor }
      image: { get_param: image }
      networks: { get_attr: [network_settings, networks] }
"""

template_value_from_nested_stack_network = """
heat_template_version: 2016-10-14
parameters:
  public_net:
    type: string
outputs:
  networks:
    value:
      - uuid: { get_param: public_net }
"""


class TestTranslation(functional_base.FunctionalTestsBase):

    def test_create_update_subnet_old_network(self):
        # Just create and update where network is translated properly.
        env = {'parameters': {'net_cidr': '11.11.11.0/24'}}
        stack_identifier = self.stack_create(
            template=template_subnet_old_network,
            environment=env)
        env = {'parameters': {'net_cidr': '11.11.12.0/24'}}
        self.update_stack(stack_identifier,
                          template=template_subnet_old_network,
                          environment=env)

    def test_create_update_translation_with_get_attr(self):
        # Check create and update successful for translation function value.
        env = {'parameters': {'net_cidr': '11.11.11.0/24'}}
        stack_identifier = self.stack_create(
            template=template_with_get_attr,
            environment=env)
        env = {'parameters': {'net_cidr': '11.11.12.0/24'}}
        self.update_stack(stack_identifier,
                          template=template_with_get_attr,
                          environment=env)

    def test_value_from_nested_stack(self):
        env = {'parameters': {
            'flavor': self.conf.minimal_instance_type,
            'image': self.conf.minimal_image_ref,
            'public_net': self.conf.fixed_network_name
        }}
        self.stack_create(
            template=template_value_from_nested_stack_main,
            environment=env,
            files={'network.yaml': template_value_from_nested_stack_network})



# Copyright 2014 Intel Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""ResourceData object."""

from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from heat.common import exception
from heat.db import api as db_api
from heat.objects import base as heat_base


class ResourceData(
    heat_base.HeatObject,
    base.VersionedObjectDictCompat,
    base.ComparableVersionedObject,
):
    fields = {
        'id': fields.IntegerField(),
        'created_at': fields.DateTimeField(read_only=True),
        'updated_at': fields.DateTimeField(nullable=True),
        'key': fields.StringField(nullable=True),
        'value': fields.StringField(nullable=True),
        'redact': fields.BooleanField(nullable=True),
        'resource_id': fields.IntegerField(),
        'decrypt_method': fields.StringField(nullable=True),
    }

    @staticmethod
    def _from_db_object(sdata, db_sdata):
        if db_sdata is None:
            return None
        for field in sdata.fields:
            sdata[field] = db_sdata[field]
        sdata.obj_reset_changes()
        return sdata

    @classmethod
    def get_all(cls, resource, *args, **kwargs):
        # this method only returns dict, so we won't use objects mechanism here
        return db_api.resource_data_get_all(resource.context,
                                            resource.id,
                                            *args,
                                            **kwargs)

    @classmethod
    def get_obj(cls, resource, key):
        raise exception.NotSupported(feature='ResourceData.get_obj')

    @classmethod
    def get_val(cls, resource, key):
        return db_api.resource_data_get(resource.context, resource.id, key)

    @classmethod
    def set(cls, resource, key, value, *args, **kwargs):
        db_data = db_api.resource_data_set(
            resource.context,
            resource.id,
            key,
            value,
            *args,
            **kwargs
        )
        return db_data

    @classmethod
    def get_by_key(cls, context, resource_id, key):
        db_rdata = db_api.resource_data_get_by_key(context, resource_id, key)
        return cls._from_db_object(cls(context), db_rdata)

    @classmethod
    def delete(cls, resource, key):
        db_api.resource_data_delete(resource.context, resource.id, key)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from keystoneauth1 import loading as ks_loading
from oslo_log import log as logging

from heat.common import exception

LOG = logging.getLogger(__name__)


def parse_auth_credential_to_dict(cred):
    """Parse credential to dict"""
    def validate(cred):
        valid_keys = ['auth_type', 'auth']
        for k in valid_keys:
            if k not in cred:
                raise ValueError('Missing key in auth information, the '
                                 'correct format contains %s.' % valid_keys)
    try:
        _cred = json.loads(cred)
    except ValueError as e:
        LOG.error('Failed to parse credential with error: %s' % e)
        raise ValueError('Failed to parse credential, please check your '
                         'Stack Credential format.')
    validate(_cred)
    return _cred


def validate_auth_plugin(auth_plugin, keystone_session):
    """Validate if this auth_plugin is valid to use."""

    try:
        auth_plugin.get_token(keystone_session)
    except Exception as e:
        # TODO(ricolin) Add heat document link for plugin information,
        # once we generated one.
        failure_reason = ("Failed to validate auth_plugin with error %s. "
                          "Please make sure the credential you provide is "
                          "correct. Also make sure the it is a valid Keystone "
                          "auth plugin type and contain in your "
                          "environment." % e)
        raise exception.AuthorizationFailure(failure_reason=failure_reason)


def get_keystone_plugin_loader(auth, keystone_session):
    cred = parse_auth_credential_to_dict(auth)
    auth_plugin = ks_loading.get_plugin_loader(
        cred.get('auth_type')).load_from_options(
            **cred.get('auth'))
    validate_auth_plugin(auth_plugin, keystone_session)
    return auth_plugin



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat_integrationtests.functional import functional_base


class SoftwareDeploymentGroupTest(functional_base.FunctionalTestsBase):
    sd_template = '''
heat_template_version: 2016-10-14

parameters:
  input:
    type: string
    default: foo_input

resources:
  config:
    type: OS::Heat::SoftwareConfig
    properties:
      group: script
      inputs:
      - name: foo

  deployment:
    type: OS::Heat::SoftwareDeploymentGroup
    properties:
      config: {get_resource: config}
      input_values:
        foo: {get_param: input}
      servers:
        '0': dummy0
        '1': dummy1
        '2': dummy2
        '3': dummy3
'''

    sd_template_with_upd_policy = '''
heat_template_version: 2016-10-14

parameters:
  input:
    type: string
    default: foo_input

resources:
  config:
    type: OS::Heat::SoftwareConfig
    properties:
      group: script
      inputs:
      - name: foo

  deployment:
    type: OS::Heat::SoftwareDeploymentGroup
    update_policy:
      rolling_update:
        max_batch_size: 2
        pause_time: 1
    properties:
      config: {get_resource: config}
      input_values:
        foo: {get_param: input}
      servers:
        '0': dummy0
        '1': dummy1
        '2': dummy2
        '3': dummy3
'''
    enable_cleanup = True

    def deployment_crud(self, template):
        stack_identifier = self.stack_create(
            template=template,
            enable_cleanup=self.enable_cleanup,
            expected_status='CREATE_IN_PROGRESS')
        self._wait_for_resource_status(
            stack_identifier, 'deployment', 'CREATE_IN_PROGRESS')

        # Wait for all deployment resources to become IN_PROGRESS, since only
        # IN_PROGRESS resources get signalled
        nested_identifier = self.assert_resource_is_a_stack(
            stack_identifier, 'deployment')
        self._wait_for_stack_status(nested_identifier, 'CREATE_IN_PROGRESS')
        self._wait_for_all_resource_status(nested_identifier,
                                           'CREATE_IN_PROGRESS')
        group_resources = self.list_group_resources(
            stack_identifier, 'deployment', minimal=False)

        self.assertEqual(4, len(group_resources))
        self._wait_for_stack_status(stack_identifier, 'CREATE_COMPLETE',
                                    signal_required=True,
                                    resources_to_signal=group_resources)

        created_group_resources = self.list_group_resources(
            stack_identifier, 'deployment', minimal=False)
        self.assertEqual(4, len(created_group_resources))
        self.check_input_values(created_group_resources, 'foo', 'foo_input')

        self.update_stack(stack_identifier,
                          template=template,
                          environment={'parameters': {'input': 'input2'}},
                          expected_status='UPDATE_IN_PROGRESS')
        nested_identifier = self.assert_resource_is_a_stack(
            stack_identifier, 'deployment')
        self._wait_for_stack_status(stack_identifier, 'UPDATE_COMPLETE',
                                    signal_required=True,
                                    resources_to_signal=group_resources)

        self.check_input_values(created_group_resources, 'foo', 'input2')

        # We explicitly test delete here, vs just via cleanup and check
        # the nested stack is gone
        self._stack_delete(stack_identifier)
        self._wait_for_stack_status(
            nested_identifier, 'DELETE_COMPLETE',
            success_on_not_found=True)

    def test_deployment_crud(self):
        self.deployment_crud(self.sd_template)

    def test_deployment_crud_with_rolling_update(self):
        self.deployment_crud(self.sd_template_with_upd_policy)

    def test_deployments_create_delete_in_progress(self):
        stack_identifier = self.stack_create(
            template=self.sd_template,
            enable_cleanup=self.enable_cleanup,
            expected_status='CREATE_IN_PROGRESS')
        self._wait_for_resource_status(
            stack_identifier, 'deployment', 'CREATE_IN_PROGRESS')
        nested_identifier = self.assert_resource_is_a_stack(
            stack_identifier, 'deployment')
        group_resources = self.list_group_resources(
            stack_identifier, 'deployment', minimal=False)

        self.assertEqual(4, len(group_resources))
        # Now test delete while the stacks are still IN_PROGRESS
        self._stack_delete(stack_identifier)
        self._wait_for_stack_status(
            nested_identifier, 'DELETE_COMPLETE',
            success_on_not_found=True)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import eventlet
from oslo_utils import timeutils

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class TestResource(resource.Resource):
    """A resource which stores the string value that was provided.

    This resource is to be used only for testing.
    It has control knobs such as 'update_replace', 'fail', 'wait_secs'.
    """

    support_status = support.SupportStatus(version='5.0.0')

    ACTION_TIMES = (
        CREATE_WAIT_SECS, UPDATE_WAIT_SECS, DELETE_WAIT_SECS
    ) = (
        'create', 'update', 'delete')

    PROPERTIES = (
        VALUE, UPDATE_REPLACE, FAIL,
        CLIENT_NAME, ENTITY_NAME,
        WAIT_SECS, ACTION_WAIT_SECS, ATTR_WAIT_SECS,
        CONSTRAINT_PROP_SECS, UPDATE_REPLACE_VALUE,
    ) = (
        'value', 'update_replace', 'fail',
        'client_name', 'entity_name',
        'wait_secs', 'action_wait_secs', 'attr_wait_secs',
        'constraint_prop_secs', 'update_replace_value',
    )

    ATTRIBUTES = (
        OUTPUT,
    ) = (
        'output',
    )

    properties_schema = {
        CONSTRAINT_PROP_SECS: properties.Schema(
            properties.Schema.NUMBER,
            _('Number value for delay during resolve constraint.'),
            default=0,
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('test_constr')
            ],
            support_status=support.SupportStatus(version='6.0.0')
        ),
        ATTR_WAIT_SECS: properties.Schema(
            properties.Schema.NUMBER,
            _('Number value for timeout during resolving output value.'),
            default=0,
            update_allowed=True,
            support_status=support.SupportStatus(version='6.0.0')
        ),
        VALUE: properties.Schema(
            properties.Schema.STRING,
            _('The input string to be stored.'),
            default='test_string',
            update_allowed=True
        ),
        UPDATE_REPLACE_VALUE: properties.Schema(
            properties.Schema.STRING,
            _('Some value that can be stored but can not be updated.'),
            support_status=support.SupportStatus(version='7.0.0')
        ),
        FAIL: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Value which can be set to fail the resource operation '
              'to test failure scenarios.'),
            update_allowed=True,
            default=False
        ),
        UPDATE_REPLACE: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Value which can be set to trigger update replace for '
              'the particular resource.'),
            update_allowed=True,
            default=False
        ),
        WAIT_SECS: properties.Schema(
            properties.Schema.NUMBER,
            _('Seconds to wait after an action (-1 is infinite).'),
            update_allowed=True,
            default=0,
        ),
        ACTION_WAIT_SECS: properties.Schema(
            properties.Schema.MAP,
            _('Options for simulating waiting.'),
            update_allowed=True,
            schema={
                CREATE_WAIT_SECS: properties.Schema(
                    properties.Schema.NUMBER,
                    _('Seconds to wait after a create. '
                      'Defaults to the global wait_secs.'),
                    update_allowed=True,
                ),
                UPDATE_WAIT_SECS: properties.Schema(
                    properties.Schema.NUMBER,
                    _('Seconds to wait after an update. '
                      'Defaults to the global wait_secs.'),
                    update_allowed=True,
                ),
                DELETE_WAIT_SECS: properties.Schema(
                    properties.Schema.NUMBER,
                    _('Seconds to wait after a delete. '
                      'Defaults to the global wait_secs.'),
                    update_allowed=True,
                ),
            }
        ),
        CLIENT_NAME: properties.Schema(
            properties.Schema.STRING,
            _('Client to poll.'),
            default='',
            update_allowed=True
        ),
        ENTITY_NAME: properties.Schema(
            properties.Schema.STRING,
            _('Client entity to poll.'),
            default='',
            update_allowed=True
        ),
    }

    attributes_schema = {
        OUTPUT: attributes.Schema(
            _('The string that was stored. This value is '
              'also available by referencing the resource.'),
            cache_mode=attributes.Schema.CACHE_NONE
        ),
    }

    def _wait_secs(self):
        secs = None
        if self.properties[self.ACTION_WAIT_SECS]:
            secs = self.properties[self.ACTION_WAIT_SECS][self.action.lower()]
        if secs is None:
            secs = self.properties[self.WAIT_SECS]
        LOG.info('%(name)s wait_secs:%(wait)s, action:%(action)s',
                 {'name': self.name,
                  'wait': secs,
                  'action': self.action.lower()})
        return secs

    def handle_create(self):
        fail_prop = self.properties.get(self.FAIL)
        if not fail_prop:
            value = self.properties.get(self.VALUE)
            self.data_set('value', value, redact=False)
            self.resource_id_set(self.physical_resource_name())

        return timeutils.utcnow(), self._wait_secs()

    def needs_replace_with_prop_diff(self, changed_properties_set,
                                     after_props, before_props):
        if self.VALUE in changed_properties_set:
            return after_props[self.UPDATE_REPLACE]

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self.properties = json_snippet.properties(self.properties_schema,
                                                  self.context)
        value = prop_diff.get(self.VALUE)
        if value:
            # emulate failure
            fail_prop = self.properties[self.FAIL]
            if not fail_prop:
                # update in place
                self.data_set('value', value, redact=False)
            return timeutils.utcnow(), self._wait_secs()
        return timeutils.utcnow(), 0

    def handle_delete(self):
        return timeutils.utcnow(), self._wait_secs()

    def check_create_complete(self, cookie):
        return self._check_status_complete(*cookie)

    def check_update_complete(self, cookie):
        return self._check_status_complete(*cookie)

    def check_delete_complete(self, cookie):
        return self._check_status_complete(*cookie)

    def _check_status_complete(self, started_at, wait_secs):
        def simulated_effort():
            client_name = self.properties[self.CLIENT_NAME]
            self.entity = self.properties[self.ENTITY_NAME]
            if client_name and self.entity:
                # Allow the user to set the value to a real resource id.
                entity_id = self.data().get('value') or self.resource_id
                try:
                    obj = getattr(self.client(name=client_name), self.entity)
                    obj.get(entity_id)
                except Exception as exc:
                    LOG.debug('%s.%s(%s) %s' % (client_name, self.entity,
                                                entity_id, str(exc)))
            else:
                # just sleep some more
                eventlet.sleep(1)

        if isinstance(started_at, str):
            started_at = timeutils.parse_isotime(started_at)

        started_at = timeutils.normalize_time(started_at)
        waited = timeutils.utcnow() - started_at
        LOG.info("Resource %(name)s waited %(waited)s/%(sec)s seconds",
                 {'name': self.name,
                  'waited': waited,
                  'sec': wait_secs})

        # wait_secs < 0 is an infinite wait time.
        if wait_secs >= 0 and waited > datetime.timedelta(seconds=wait_secs):
            fail_prop = self.properties[self.FAIL]
            if fail_prop and self.action != self.DELETE:
                raise ValueError("Test Resource failed %s" % self.name)
            return True

        simulated_effort()
        return False

    def _resolve_attribute(self, name):
        eventlet.sleep(self.properties[self.ATTR_WAIT_SECS])
        if name == self.OUTPUT:
            return self.data().get('value')


def resource_mapping():
    return {
        'OS::Heat::TestResource': TestResource,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation


class OctaviaQuota(resource.Resource):
    """A resource for creating Octavia quotas.

    Ocatavia Quota is used to manage operational limits for Octavia. Currently,
    this resource can manage Octavia's quotas for:

        - healthmonitor
        - listener
        - loadbalancer
        - pool
        - member

    Note that default octavia security policy usage of this resource
    is limited to being used by administrators only. Administrators should be
    careful to create only one Octavia Quota resource per project, otherwise
    it will be hard for them to manage the quota properly.
    """

    support_status = support.SupportStatus(version='14.0.0')

    default_client_name = 'octavia'

    entity = 'quotas'

    PROPERTIES = (
        PROJECT, HEALTHMONITOR, LISTENER, LOADBALANCER,
        POOL, MEMBER
    ) = (
        'project', 'healthmonitor', 'listener', 'loadbalancer',
        'pool', 'member'
    )

    properties_schema = {
        PROJECT: properties.Schema(
            properties.Schema.STRING,
            _('Name or id of the project to set the quota for.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('keystone.project')
            ]
        ),
        HEALTHMONITOR: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of healthmonitors. '
              'Setting the value to -1 removes the limit.'),
            constraints=[
                constraints.Range(min=-1),
            ],
            update_allowed=True
        ),
        LISTENER: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of listeners. '
              'Setting the value to -1 removes the limit.'),
            constraints=[
                constraints.Range(min=-1),
            ],
            update_allowed=True
        ),
        LOADBALANCER: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of load balancers. '
              'Setting the value to -1 removes the limit.'),
            constraints=[
                constraints.Range(min=-1),
            ],
            update_allowed=True
        ),
        POOL: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of pools. '
              'Setting the value to -1 removes the limit.'),
            constraints=[
                constraints.Range(min=-1),
            ],
            update_allowed=True
        ),
        MEMBER: properties.Schema(
            properties.Schema.INTEGER,
            _('Quota for the number of m. '
              'Setting the value to -1 removes the limit.'),
            constraints=[
                constraints.Range(min=-1),
            ],
            update_allowed=True
        ),
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.PROJECT],
                client_plugin=self.client_plugin('keystone'),
                finder='get_project_id')
        ]

    def handle_create(self):
        self._set_quota()
        self.resource_id_set(self.physical_resource_name())

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self._set_quota(json_snippet.properties(self.properties_schema,
                                                self.context))

    def _set_quota(self, props=None):
        if props is None:
            props = self.properties

        kwargs = dict((k, v) for k, v in props.items()
                      if k != self.PROJECT and v is not None)
        self.client().quota_set(props.get(self.PROJECT), **kwargs)

    def handle_delete(self):
        self.client().quota_reset(self.properties[self.PROJECT])

    def validate(self):
        super(OctaviaQuota, self).validate()
        if sum(1 for p in self.properties.values() if p is not None) <= 1:
            raise exception.PropertyUnspecifiedError(
                *sorted(set(self.PROPERTIES) - {self.PROJECT}))


def resource_mapping():
    return {
        'OS::Octavia::Quota': OctaviaQuota
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine.hot import parameters
from heat.engine import properties
from heat.engine import resource
from heat.engine import support


class Value(resource.Resource):
    """A resource which exposes its value property as an attribute.

    This is useful for exposing a value that is a simple manipulation
    of other template parameters and/or other resources.
    """

    support_status = support.SupportStatus(version='7.0.0')

    PROPERTIES = (
        VALUE, TYPE,
    ) = (
        'value', constraints.Schema.TYPE,
    )

    ATTRIBUTES = (
        VALUE_ATTR,
    ) = (
        'value',
    )

    properties_schema = {
        VALUE: properties.Schema(
            properties.Schema.ANY,
            _('The expression to generate the "value" attribute.'),
            required=True,
            update_allowed=True,
        ),
        TYPE: properties.Schema(
            properties.Schema.STRING,
            _('The type of the "value" property.'),
            constraints=[constraints.AllowedValues(
                parameters.HOTParamSchema.TYPES)],
            update_allowed=True,
        ),
    }

    attributes_schema = {
        VALUE_ATTR: attributes.Schema(
            _('The value generated by this resource\'s properties "value" '
              'expression, with type determined from the properties "type".'),
        ),
    }

    def _resolve_attribute(self, name):
        props = self.frozen_definition().properties(self.properties_schema,
                                                    self.context)
        if name == self.VALUE_ATTR:
            return props[self.VALUE]

    def handle_create(self):
        self.resource_id_set(self.physical_resource_name())

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        # allow update, no replace necessary (do not throw UpdateReplace).
        # the resource properties are updated appropriately in parent class.
        pass

    def reparse(self, *args, **kwargs):
        super(Value, self).reparse(*args, **kwargs)
        value_type = self.properties[self.TYPE]
        if value_type is None:
            # We don't know what type the value is, anything goes
            return

        param_type_map = {
            parameters.HOTParamSchema.STRING: constraints.Schema.STRING,
            parameters.HOTParamSchema.NUMBER: constraints.Schema.NUMBER,
            parameters.HOTParamSchema.LIST: constraints.Schema.LIST,
            parameters.HOTParamSchema.MAP: constraints.Schema.MAP,
            parameters.HOTParamSchema.BOOLEAN: constraints.Schema.BOOLEAN,
        }
        if value_type not in param_type_map:
            # the self.TYPE value is not valid, the error will become
            # apparent appropriately when its constraints are checked.
            return
        # We know the value's type, update value's property schema accordingly
        self.properties.props[self.VALUE] = properties.Property(
            properties.Schema(
                param_type_map[value_type],
                _('The expression to generate the "value" attribute.'),
                required=True,
                update_allowed=True,
            ),
            self.VALUE)


def resource_mapping():
    return {
        'OS::Heat::Value': Value,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine.resources.aws.ec2 import vpc
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support


class RouteTable(resource.Resource):

    support_status = support.SupportStatus(version='2014.1')

    PROPERTIES = (
        VPC_ID, TAGS,
    ) = (
        'VpcId', 'Tags',
    )

    _TAG_KEYS = (
        TAG_KEY, TAG_VALUE,
    ) = (
        'Key', 'Value',
    )

    properties_schema = {
        VPC_ID: properties.Schema(
            properties.Schema.STRING,
            _('VPC ID for where the route table is created.'),
            required=True
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            schema=properties.Schema(
                properties.Schema.MAP,
                _('List of tags to be attached to this resource.'),
                schema={
                    TAG_KEY: properties.Schema(
                        properties.Schema.STRING,
                        required=True
                    ),
                    TAG_VALUE: properties.Schema(
                        properties.Schema.STRING,
                        required=True
                    ),
                },
                implemented=False,
            )
        ),
    }

    default_client_name = 'neutron'

    def handle_create(self):
        client = self.client()
        props = {'name': self.physical_resource_name()}
        router = client.create_router({'router': props})['router']
        self.resource_id_set(router['id'])

    def check_create_complete(self, *args):
        client = self.client()
        attributes = client.show_router(
            self.resource_id)['router']
        if not neutron.NeutronResource.is_built(attributes):
            return False

        network_id = self.properties.get(self.VPC_ID)
        default_router = vpc.VPC.router_for_vpc(client, network_id)
        if default_router and default_router.get('external_gateway_info'):
            # the default router for the VPC is connected
            # to the external router, so do it for this too.
            external_network_id = default_router[
                'external_gateway_info']['network_id']
            client.add_gateway_router(self.resource_id, {
                'network_id': external_network_id})
        return True

    def handle_delete(self):
        client = self.client()

        router_id = self.resource_id
        with self.client_plugin().ignore_not_found:
            client.delete_router(router_id)

        # just in case this router has been added to a gateway, remove it
        with self.client_plugin().ignore_not_found:
            client.remove_gateway_router(router_id)


class SubnetRouteTableAssociation(resource.Resource):

    PROPERTIES = (
        ROUTE_TABLE_ID, SUBNET_ID,
    ) = (
        'RouteTableId', 'SubnetId',
    )

    properties_schema = {
        ROUTE_TABLE_ID: properties.Schema(
            properties.Schema.STRING,
            _('Route table ID.'),
            required=True
        ),
        SUBNET_ID: properties.Schema(
            properties.Schema.STRING,
            _('Subnet ID.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('neutron.subnet')
            ]
        ),
    }

    default_client_name = 'neutron'

    def handle_create(self):
        client = self.client()
        subnet_id = self.properties.get(self.SUBNET_ID)

        router_id = self.properties.get(self.ROUTE_TABLE_ID)

        # remove the default router association for this subnet.
        with self.client_plugin().ignore_not_found:
            previous_router = self._router_for_subnet(subnet_id)
            if previous_router:
                client.remove_interface_router(
                    previous_router['id'],
                    {'subnet_id': subnet_id})

        client.add_interface_router(
            router_id, {'subnet_id': subnet_id})

    def _router_for_subnet(self, subnet_id):
        client = self.client()
        subnet = client.show_subnet(
            subnet_id)['subnet']
        network_id = subnet['network_id']
        return vpc.VPC.router_for_vpc(client, network_id)

    def handle_delete(self):
        client = self.client()
        subnet_id = self.properties.get(self.SUBNET_ID)

        router_id = self.properties.get(self.ROUTE_TABLE_ID)

        with self.client_plugin().ignore_not_found:
            client.remove_interface_router(router_id, {
                'subnet_id': subnet_id})

        # add back the default router
        with self.client_plugin().ignore_not_found:
            default_router = self._router_for_subnet(subnet_id)
            if default_router:
                client.add_interface_router(
                    default_router['id'], {'subnet_id': subnet_id})


def resource_mapping():
    return {
        'AWS::EC2::RouteTable': RouteTable,
        'AWS::EC2::SubnetRouteTableAssociation': SubnetRouteTableAssociation,
    }



#
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Heat exception subclasses"""

import sys

from oslo_log import log as logging
from oslo_utils import excutils

from heat.common.i18n import _

_FATAL_EXCEPTION_FORMAT_ERRORS = False


LOG = logging.getLogger(__name__)


# TODO(kanagaraj-manickam): Expose this to user via REST API
ERROR_CODE_MAP = {
    '99001': _("Service %(service_name)s is not available for resource "
               "type %(resource_type)s, reason: %(reason)s")
}


class HeatException(Exception):
    """Base Heat Exception.

    To correctly use this class, inherit from it and define a 'msg_fmt'
    property. That msg_fmt will get formatted with the keyword arguments
    provided to the constructor.
    """
    message = _("An unknown exception occurred.")
    # error_code helps to provide an unique number for a given exception
    # and is encoded in XXYYY format.
    # Here, XX - For each of the entity type like stack, resource, etc
    # an unique number will be provided. All exceptions for a entity will
    # have same XX code.
    # YYY - Specific error code for a given exception.
    error_code = None

    safe = True

    def __init__(self, **kwargs):
        self.kwargs = kwargs

        if self.error_code in ERROR_CODE_MAP:
            self.msg_fmt = ERROR_CODE_MAP[self.error_code]

        try:
            self.message = self.msg_fmt % kwargs
        except KeyError:
            with excutils.save_and_reraise_exception(
                    reraise=_FATAL_EXCEPTION_FORMAT_ERRORS):
                # kwargs doesn't match a variable in the message
                # log the issue and the kwargs
                LOG.exception('Exception in string format operation')
                for name, value in kwargs.items():
                    LOG.error("%(name)s: %(value)s",
                              {'name': name, 'value': value})  # noqa

        if self.error_code:
            self.message = 'HEAT-E%s %s' % (self.error_code, self.message)

    def __str__(self):
        return self.message

    def __deepcopy__(self, memo):
        return self.__class__(**self.kwargs)


class MissingCredentialError(HeatException):
    msg_fmt = _("Missing required credential: %(required)s")


class AuthorizationFailure(HeatException):
    msg_fmt = _("Authorization failed.%(failure_reason)s")

    def __init__(self, failure_reason=""):
        if failure_reason != "":
            # Add a space to make message more readable
            failure_reason = " " + failure_reason
        super(AuthorizationFailure, self).__init__(
            failure_reason=failure_reason)


class NotAuthenticated(HeatException):
    msg_fmt = _("You are not authenticated.")


class Forbidden(HeatException):
    msg_fmt = _("You are not authorized to use %(action)s.")

    def __init__(self, action='this action'):
        super(Forbidden, self).__init__(action=action)


# NOTE(bcwaldon): here for backwards-compatibility, need to deprecate.
class NotAuthorized(Forbidden):
    msg_fmt = _("You are not authorized to complete this action.")


class Invalid(HeatException):
    msg_fmt = _("Data supplied was not valid: %(reason)s")


class UserParameterMissing(HeatException):
    msg_fmt = _("The Parameter (%(key)s) was not provided.")


class UnknownUserParameter(HeatException):
    msg_fmt = _("The Parameter (%(key)s) was not defined in template.")


class InvalidTemplateVersion(HeatException):
    msg_fmt = _("The template version is invalid: %(explanation)s")


class InvalidTemplateSection(HeatException):
    msg_fmt = _("The template section is invalid: %(section)s")


class ImmutableParameterModified(HeatException):
    msg_fmt = _("The following parameters are immutable and may not be "
                "updated: %(keys)s")

    def __init__(self, *args, **kwargs):
        if args:
            kwargs.update({'keys': ", ".join(args)})
        super(ImmutableParameterModified, self).__init__(**kwargs)


class InvalidMergeStrategyForParam(HeatException):
    msg_fmt = _("Invalid merge strategy '%(strategy)s' for "
                "parameter '%(param)s'.")


class ConflictingMergeStrategyForParam(HeatException):
    msg_fmt = _("Conflicting merge strategy '%(strategy)s' for "
                "parameter '%(param)s' in file '%(env_file)s'.")


class InvalidTemplateAttribute(HeatException):
    msg_fmt = _("The Referenced Attribute (%(resource)s %(key)s)"
                " is incorrect.")


class InvalidTemplateReference(HeatException):
    msg_fmt = _('The specified reference "%(resource)s" (in %(key)s)'
                ' is incorrect.')


class TemplateOutputError(HeatException):
    msg_fmt = _('Error in %(resource)s output %(attribute)s: %(message)s')


class InvalidEncryptionKey(HeatException):
    msg_fmt = _('Can not decrypt data with the auth_encryption_key'
                ' in heat config.')


class InvalidExternalResourceDependency(HeatException):
    msg_fmt = _("Invalid dependency with external %(resource_type)s "
                "resource: %(external_id)s")


class EntityNotFound(HeatException):
    msg_fmt = _("The %(entity)s (%(name)s) could not be found.")

    def __init__(self, entity=None, name=None, **kwargs):
        self.entity = entity
        self.name = name
        super(EntityNotFound, self).__init__(entity=entity, name=name,
                                             **kwargs)


class PhysicalResourceExists(HeatException):
    msg_fmt = _("The physical resource for (%(name)s) exists.")


class PhysicalResourceNameAmbiguity(HeatException):
    msg_fmt = _(
        "Multiple physical resources were found with name (%(name)s).")


class PhysicalResourceIDAmbiguity(HeatException):
    msg_fmt = _(
        "Multiple resources were found with the physical ID (%(phys_id)s).")


class InvalidTenant(HeatException):
    msg_fmt = _("Searching Tenant %(target)s "
                "from Tenant %(actual)s forbidden.")


class StackExists(HeatException):
    msg_fmt = _("The Stack (%(stack_name)s) already exists.")


class HeatExceptionWithPath(HeatException):
    msg_fmt = _("%(error)s%(path)s%(message)s")

    def __init__(self, error=None, path=None, message=None):
        self.error = error or ''
        self.path = []
        if path is not None:
            if isinstance(path, list):
                self.path = path
            elif isinstance(path, str):
                self.path = [path]

        result_path = ''
        for path_item in self.path:
            if isinstance(path_item, int) or path_item.isdigit():
                result_path += '[%s]' % path_item
            elif len(result_path) > 0:
                result_path += '.%s' % path_item
            else:
                result_path = path_item

        self.error_message = message or ''
        super(HeatExceptionWithPath, self).__init__(
            error=('%s: ' % self.error if self.error != '' else ''),
            path=('%s: ' % result_path if len(result_path) > 0 else ''),
            message=self.error_message
        )


class StackValidationFailed(HeatExceptionWithPath):
    def __init__(self, error=None, path=None, message=None,
                 resource=None):
        if path is None:
            path = []
        elif isinstance(path, str):
            path = [path]

        if resource is not None and not path:
            path = [resource.stack.t.get_section_name(
                resource.stack.t.RESOURCES), resource.name]
        if isinstance(error, Exception):
            if isinstance(error, StackValidationFailed):
                str_error = error.error
                message = error.error_message
                path = path + error.path
                # This is a hack to avoid the py3 (chained exception)
                # json serialization circular reference error from
                # oslo.messaging.
                self.args = error.args
            else:
                str_error = str(type(error).__name__)
                message = str(error)
        else:
            str_error = error

        super(StackValidationFailed, self).__init__(error=str_error, path=path,
                                                    message=message)


class InvalidSchemaError(HeatException):
    msg_fmt = _("%(message)s")


class ResourceNotFound(EntityNotFound):
    msg_fmt = _("The Resource (%(resource_name)s) could not be found "
                "in Stack %(stack_name)s.")


class SnapshotNotFound(EntityNotFound):
    msg_fmt = _("The Snapshot (%(snapshot)s) for Stack (%(stack)s) "
                "could not be found.")


class InvalidGlobalResource(HeatException):
    msg_fmt = _("There was an error loading the definition of the global "
                "resource type %(type_name)s.")


class ResourceTypeUnavailable(HeatException):
    error_code = '99001'


class InvalidBreakPointHook(HeatException):
    msg_fmt = _("%(message)s")


class InvalidRestrictedAction(HeatException):
    msg_fmt = _("%(message)s")


class ResourceNotAvailable(HeatException):
    msg_fmt = _("The Resource (%(resource_name)s) is not available.")


class ClientNotAvailable(HeatException):
    msg_fmt = _("The client (%(client_name)s) is not available.")


class ResourceFailure(HeatExceptionWithPath):
    def __init__(self, exception_or_error, resource, action=None):
        self.resource = resource
        self.action = action
        if action is None and resource is not None:
            self.action = resource.action
        path = []
        res_path = []
        if resource is not None:
            res_path = [resource.stack.t.get_section_name('resources'),
                        resource.name]

        if isinstance(exception_or_error, Exception):
            if isinstance(exception_or_error, ResourceFailure):
                self.exc = exception_or_error.exc
                error = exception_or_error.error
                message = exception_or_error.error_message
                path = exception_or_error.path
            else:
                self.exc = exception_or_error
                error = str(type(self.exc).__name__)
                message = str(self.exc)
                path = res_path
        else:
            self.exc = None
            res_failed = 'Resource %s failed: ' % self.action.upper()
            if res_failed in exception_or_error:
                (error, message, new_path) = self._from_status_reason(
                    exception_or_error)
                path = res_path + new_path
            else:
                path = res_path
                error = None
                message = exception_or_error

        super(ResourceFailure, self).__init__(error=error, path=path,
                                              message=message)

    def _from_status_reason(self, status_reason):
        """Split the status_reason up into parts.

        Given the following status_reason:
        "Resource DELETE failed: Exception : resources.AResource: foo"

        we are going to return:
        ("Exception", "resources.AResource", "foo")
        """
        parsed = [sp.strip() for sp in status_reason.split(':')]
        if len(parsed) >= 4:
            error = parsed[1]
            message = ': '.join(parsed[3:])
            path = parsed[2].split('.')
        else:
            error = ''
            message = status_reason
            path = []
        return (error, message, path)


class NotSupported(HeatException):
    msg_fmt = _("%(feature)s is not supported.")


class ResourceActionNotSupported(HeatException):
    msg_fmt = _("%(action)s is not supported for resource.")


class ResourceActionRestricted(HeatException):
    msg_fmt = _("%(action)s is restricted for resource.")


class ResourcePropertyConflict(HeatException):
    msg_fmt = _('Cannot define the following properties '
                'at the same time: %(props)s.')

    def __init__(self, *args, **kwargs):
        if args:
            kwargs.update({'props': ", ".join(args)})
        super(ResourcePropertyConflict, self).__init__(**kwargs)


class ResourcePropertyDependency(HeatException):
    msg_fmt = _('%(prop1)s cannot be specified without %(prop2)s.')


class ResourcePropertyValueDependency(HeatException):
    msg_fmt = _('%(prop1)s property should only be specified '
                'for %(prop2)s with value %(value)s.')


class PropertyUnspecifiedError(HeatException):
    msg_fmt = _('At least one of the following properties '
                'must be specified: %(props)s.')

    def __init__(self, *args, **kwargs):
        if args:
            kwargs.update({'props': ", ".join(args)})
        super(PropertyUnspecifiedError, self).__init__(**kwargs)


# Do not reference this here - in the future it will move back to its
# correct (and original) location in heat.engine.resource. Reference it as
# heat.engine.resource.UpdateReplace instead.
class UpdateReplace(Exception):
    """Raised when resource update requires replacement."""
    def __init__(self, resource_name='Unknown'):
        msg = _("The Resource %s requires replacement.") % resource_name
        super(Exception, self).__init__(str(msg))


class ResourceUnknownStatus(HeatException):
    msg_fmt = _('%(result)s - Unknown status %(resource_status)s due to '
                '"%(status_reason)s"')

    def __init__(self, result=_('Resource failed'),
                 status_reason=_('Unknown'), **kwargs):
        super(ResourceUnknownStatus, self).__init__(
            result=result, status_reason=status_reason, **kwargs)


class ResourceInError(HeatException):
    msg_fmt = _('Went to status %(resource_status)s '
                'due to "%(status_reason)s"')

    def __init__(self, status_reason=_('Unknown'), **kwargs):
        super(ResourceInError, self).__init__(status_reason=status_reason,
                                              **kwargs)


class UpdateInProgress(Exception):
    def __init__(self, resource_name='Unknown'):
        msg = _("The resource %s is already being updated.") % resource_name
        super(Exception, self).__init__(str(msg))


class HTTPExceptionDisguise(Exception):
    """Disguises HTTP exceptions.

    They can be handled by the webob fault application in the wsgi pipeline.
    """

    safe = True

    def __init__(self, exception):
        self.exc = exception
        self.tb = sys.exc_info()[2]


class EgressRuleNotAllowed(HeatException):
    msg_fmt = _("Egress rules are only allowed when "
                "Neutron is used and the 'VpcId' property is set.")


class Error(HeatException):
    msg_fmt = "%(message)s"

    def __init__(self, msg):
        super(Error, self).__init__(message=msg)


class NotFound(HeatException):
    def __init__(self, msg_fmt=_('Not found')):
        self.msg_fmt = msg_fmt
        super(NotFound, self).__init__()


class InvalidContentType(HeatException):
    msg_fmt = _("Invalid content type %(content_type)s")


class RequestLimitExceeded(HeatException):
    msg_fmt = _('Request limit exceeded: %(message)s')


class DownloadLimitExceeded(HeatException):
    msg_fmt = _('Permissible download limit exceeded: %(message)s')


class StackResourceLimitExceeded(HeatException):
    msg_fmt = _('Maximum resources per stack exceeded.')


class ActionInProgress(HeatException):
    msg_fmt = _("Stack %(stack_name)s already has an action (%(action)s) "
                "in progress.")


class ActionNotComplete(HeatException):
    msg_fmt = _("Stack %(stack_name)s has an action (%(action)s) "
                "in progress or failed state.")


class StopActionFailed(HeatException):
    msg_fmt = _("Failed to stop stack (%(stack_name)s) on other engine "
                "(%(engine_id)s)")


class EventSendFailed(HeatException):
    msg_fmt = _("Failed to send message to stack (%(stack_name)s) "
                "on other engine (%(engine_id)s)")


class InterfaceAttachFailed(HeatException):
    msg_fmt = _("Failed to attach interface (%(port)s) "
                "to server (%(server)s)")


class InterfaceDetachFailed(HeatException):
    msg_fmt = _("Failed to detach interface (%(port)s) "
                "from server (%(server)s)")


class UnsupportedObjectError(HeatException):
    msg_fmt = _('Unsupported object type %(objtype)s')


class OrphanedObjectError(HeatException):
    msg_fmt = _('Cannot call %(method)s on orphaned %(objtype)s object')


class IncompatibleObjectVersion(HeatException):
    msg_fmt = _('Version %(objver)s of %(objname)s is not supported')


class ObjectActionError(HeatException):
    msg_fmt = _('Object action %(action)s failed because: %(reason)s')


class ReadOnlyFieldError(HeatException):
    msg_fmt = _('Cannot modify readonly field %(field)s')


class ConcurrentTransaction(HeatException):
    msg_fmt = _('Concurrent transaction for %(action)s')


class ObjectFieldInvalid(HeatException):
    msg_fmt = _('Field %(field)s of %(objname)s is not an instance of Field')


class KeystoneServiceNameConflict(HeatException):
    msg_fmt = _("Keystone has more than one service with same name "
                "%(service)s. Please use service id instead of name")


class SIGHUPInterrupt(HeatException):
    msg_fmt = _("System SIGHUP signal received.")


class InvalidServiceVersion(HeatException):
    msg_fmt = _("Invalid service %(service)s version %(version)s")


class InvalidTemplateVersions(HeatException):
    msg_fmt = _('A template version alias %(version)s was added for a '
                'template class that has no official YYYY-MM-DD version.')


class UnableToAutoAllocateNetwork(HeatException):
    msg_fmt = _('Unable to automatically allocate a network: %(message)s')


class CircularDependencyException(HeatException):
    msg_fmt = _("Circular Dependency Found: %(cycle)s")



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat.engine import node_data

from heat.tests import common


def make_test_data():
    return {
        'id': 42,
        'name': 'foo',
        'reference_id': 'foo-000000',
        'attrs': {
            'foo': 'bar',
            ('foo', 'bar', 'baz'): 'quux',
            ('blarg', 'wibble'): 'foo',
        },
        'action': 'CREATE',
        'status': 'COMPLETE',
        'uuid': '000000-0000-0000-0000000',
    }


def make_test_node():
    return node_data.NodeData.from_dict(make_test_data())


class NodeDataTest(common.HeatTestCase):
    def test_round_trip(self):
        in_dict = make_test_data()
        self.assertEqual(in_dict,
                         node_data.NodeData.from_dict(in_dict).as_dict())

    def test_resource_key(self):
        nd = make_test_node()
        self.assertEqual(42, nd.primary_key)

    def test_resource_name(self):
        nd = make_test_node()
        self.assertEqual('foo', nd.name)

    def test_action(self):
        nd = make_test_node()
        self.assertEqual('CREATE', nd.action)

    def test_status(self):
        nd = make_test_node()
        self.assertEqual('COMPLETE', nd.status)

    def test_refid(self):
        nd = make_test_node()
        self.assertEqual('foo-000000', nd.reference_id())

    def test_all_attrs(self):
        nd = make_test_node()
        self.assertEqual({'foo': 'bar'}, nd.attributes())

    def test_attr(self):
        nd = make_test_node()
        self.assertEqual('bar', nd.attribute('foo'))

    def test_path_attr(self):
        nd = make_test_node()
        self.assertEqual('quux', nd.attribute(('foo', 'bar', 'baz')))

    def test_attr_names(self):
        nd = make_test_node()
        self.assertEqual({'foo', 'blarg'}, set(nd.attribute_names()))



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_db import exception

from heat.engine import sync_point
from heat.tests import common
from heat.tests.engine import tools
from heat.tests import utils


class SyncPointTestCase(common.HeatTestCase):
    def setUp(self):
        super(SyncPointTestCase, self).setUp()
        self.dummy_event = mock.MagicMock()
        self.dummy_event.ready.return_value = False

    def test_sync_waiting(self):
        ctx = utils.dummy_context()
        stack = tools.get_stack('test_stack', utils.dummy_context(),
                                template=tools.string_template_five,
                                convergence=True)
        stack.converge_stack(stack.t, action=stack.CREATE)
        resource = stack['C']
        graph = stack.convergence_dependencies.graph()

        sender = (4, True)
        mock_callback = mock.Mock()
        sync_point.sync(ctx, resource.id, stack.current_traversal, True,
                        mock_callback, set(graph[(resource.id, True)]),
                        {sender: None})
        updated_sync_point = sync_point.get(ctx, resource.id,
                                            stack.current_traversal, True)
        input_data = sync_point.deserialize_input_data(
            updated_sync_point.input_data)
        self.assertEqual({sender: None}, input_data)
        self.assertFalse(mock_callback.called)

    def test_sync_non_waiting(self):
        ctx = utils.dummy_context()
        stack = tools.get_stack('test_stack', utils.dummy_context(),
                                template=tools.string_template_five,
                                convergence=True)
        stack.converge_stack(stack.t, action=stack.CREATE)
        resource = stack['A']
        graph = stack.convergence_dependencies.graph()

        sender = (3, True)
        mock_callback = mock.Mock()
        sync_point.sync(ctx, resource.id, stack.current_traversal, True,
                        mock_callback, set(graph[(resource.id, True)]),
                        {sender: None})
        updated_sync_point = sync_point.get(ctx, resource.id,
                                            stack.current_traversal, True)
        input_data = sync_point.deserialize_input_data(
            updated_sync_point.input_data)
        self.assertEqual({sender: None}, input_data)
        self.assertTrue(mock_callback.called)

    def test_serialize_input_data(self):
        res = sync_point.serialize_input_data({(3, 8): None})
        self.assertEqual({'input_data': {'tuple:(3, 8)': None}}, res)

    @mock.patch('heat.engine.sync_point.update_input_data', return_value=None)
    @mock.patch('time.sleep', side_effect=exception.DBError)
    def sync_with_sleep(self, ctx, stack, mock_sleep_time, mock_uid):
        resource = stack['C']
        graph = stack.convergence_dependencies.graph()

        mock_callback = mock.Mock()
        sender = (3, True)
        self.assertRaises(exception.DBError, sync_point.sync, ctx, resource.id,
                          stack.current_traversal, True, mock_callback,
                          set(graph[(resource.id, True)]), {sender: None})
        return mock_sleep_time

    def test_sync_with_time_throttle(self):
        ctx = utils.dummy_context()
        stack = tools.get_stack('test_stack', utils.dummy_context(),
                                template=tools.string_template_five,
                                convergence=True)
        stack.converge_stack(stack.t, action=stack.CREATE)
        mock_sleep_time = self.sync_with_sleep(ctx, stack)
        self.assertTrue(mock_sleep_time.called)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception as heat_exception
from heat.engine.clients.os import monasca as client_plugin
from heat.tests import common
from heat.tests import utils


class MonascaNotificationConstraintTest(common.HeatTestCase):
    def test_expected_exceptions(self):
        self.assertEqual(
            (heat_exception.EntityNotFound,),
            client_plugin.MonascaNotificationConstraint.expected_exceptions,
            "MonascaNotificationConstraint expected exceptions error")

    def test_constraint(self):
        constraint = client_plugin.MonascaNotificationConstraint()
        client_mock = mock.MagicMock()
        client_plugin_mock = mock.MagicMock()
        client_plugin_mock.get_notification.return_value = None
        client_mock.client_plugin.return_value = client_plugin_mock

        self.assertIsNone(constraint.validate_with_client(client_mock,
                                                          'notification_1'))
        client_plugin_mock.get_notification.assert_called_once_with(
            'notification_1')


class MonascaClientPluginTest(common.HeatTestCase):
    def test_client(self):
        context = utils.dummy_context()
        plugin = context.clients.client_plugin('monasca')
        client = plugin.client()
        self.assertIsNotNone(client.metrics)

    def test_client_uses_session(self):
        context = mock.MagicMock()
        monasca_client = client_plugin.MonascaClientPlugin(context=context)
        self.assertIsNotNone(monasca_client._create())


class MonascaClientPluginNotificationTest(common.HeatTestCase):

    sample_uuid = '477e8273-60a7-4c41-b683-fdb0bc7cd152'
    sample_name = 'test-notification'

    def _get_mock_notification(self):
        notification = dict()
        notification['id'] = self.sample_uuid
        notification['name'] = self.sample_name
        return notification

    def setUp(self):
        super(MonascaClientPluginNotificationTest, self).setUp()
        self._client = mock.MagicMock()
        self.client_plugin = client_plugin.MonascaClientPlugin(
            context=mock.MagicMock()
        )

    @mock.patch.object(client_plugin.MonascaClientPlugin, 'client')
    def test_get_notification(self, client_monasca):
        mock_notification = self._get_mock_notification()
        self._client.notifications.get.return_value = mock_notification
        client_monasca.return_value = self._client

        self.assertEqual(self.sample_uuid,
                         self.client_plugin.get_notification(
                             self.sample_uuid))
        self._client.notifications.get.assert_called_once_with(
            notification_id=self.sample_uuid)

    @mock.patch.object(client_plugin.MonascaClientPlugin, 'client')
    def test_get_notification_not_found(self, client_monasca):
        self._client.notifications.get.side_effect = (
            client_plugin.monasca_exc.NotFound)
        client_monasca.return_value = self._client

        ex = self.assertRaises(heat_exception.EntityNotFound,
                               self.client_plugin.get_notification,
                               self.sample_uuid)
        msg = ("The Monasca Notification (%(name)s) could not be found." %
               {'name': self.sample_uuid})
        self.assertEqual(msg, str(ex))
        self._client.notifications.get.assert_called_once_with(
            notification_id=self.sample_uuid)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat_integrationtests.functional import functional_base

# Simple stack
test_template = {
    'heat_template_version': '2013-05-23',
    'resources': {
        'test1': {
            'type': 'OS::Heat::TestResource',
            'properties': {
                'value': 'Test1'
            }
        }
    }
}

# Nested stack
rsg_template = {
    'heat_template_version': '2013-05-23',
    'resources': {
        'random_group': {
            'type': 'OS::Heat::ResourceGroup',
            'properties': {
                'count': 2,
                'resource_def': {
                    'type': 'OS::Heat::RandomString',
                    'properties': {
                        'length': 30,
                        'salt': 'initial'
                    }
                }
            }
        }
    }
}


class AdminActionsTest(functional_base.FunctionalTestsBase):

    def setUp(self):
        super(AdminActionsTest, self).setUp()
        if not self.conf.admin_username or not self.conf.admin_password:
            self.skipTest('No admin creds found, skipping')

    def create_stack_setup_admin_client(self, template=test_template):
        # Create the stack with the default user
        self.stack_identifier = self.stack_create(template=template)

        # Setup admin clients
        self.setup_clients_for_admin()

    def test_admin_simple_stack_actions(self):
        self.create_stack_setup_admin_client()

        updated_template = test_template.copy()
        props = updated_template['resources']['test1']['properties']
        props['value'] = 'new_value'

        # Update, suspend and resume stack
        self.update_stack(self.stack_identifier,
                          template=updated_template)
        self.stack_suspend(self.stack_identifier)
        self.stack_resume(self.stack_identifier)

        # List stack resources
        initial_resources = {'test1': 'OS::Heat::TestResource'}
        self.assertEqual(initial_resources,
                         self.list_resources(self.stack_identifier))
        # Delete stack
        self._stack_delete(self.stack_identifier)

    def test_admin_complex_stack_actions(self):
        self.create_stack_setup_admin_client(template=rsg_template)

        updated_template = rsg_template.copy()
        props = updated_template['resources']['random_group']['properties']
        props['count'] = 3

        # Update, suspend and resume stack
        self.update_stack(self.stack_identifier,
                          template=updated_template)
        self.stack_suspend(self.stack_identifier)
        self.stack_resume(self.stack_identifier)

        # List stack resources
        resources = {'random_group': 'OS::Heat::ResourceGroup'}
        self.assertEqual(resources,
                         self.list_resources(self.stack_identifier))
        # Delete stack
        self._stack_delete(self.stack_identifier)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from designateclient import client
from designateclient import exceptions

from heat.common import exception as heat_exception
from heat.engine.clients import client_plugin
from heat.engine import constraints

CLIENT_NAME = 'designate'


class DesignateClientPlugin(client_plugin.ClientPlugin):

    exceptions_module = [exceptions]

    service_types = [DNS] = ['dns']

    def _create(self):
        endpoint_type = self._get_client_option(CLIENT_NAME, 'endpoint_type')
        return client.Client(version='2',
                             session=self.context.keystone_session,
                             endpoint_type=endpoint_type,
                             service_type=self.DNS,
                             region_name=self._get_region_name())

    def is_not_found(self, ex):
        return isinstance(ex, exceptions.NotFound)

    def get_zone_id(self, zone_id_or_name):
        client = self.client()
        try:
            zone_obj = client.zones.get(zone_id_or_name)
            return zone_obj['id']
        except exceptions.NotFound:
            zones = client.zones.list(criterion=dict(name=zone_id_or_name))
            if len(zones) == 1:
                return zones[0]['id']

        raise heat_exception.EntityNotFound(entity='Designate Zone',
                                            name=zone_id_or_name)


class DesignateZoneConstraint(constraints.BaseCustomConstraint):
    resource_client_name = CLIENT_NAME
    resource_getter_name = 'get_zone_id'



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import io
import random as random_module
import string


# NOTE(pas-ha) Heat officially supports only POSIX::Linux platform
# where os.urandom() and random.SystemRandom() are available
random = random_module.SystemRandom()


CHARACTER_CLASSES = (
    LETTERS_DIGITS, LETTERS, LOWERCASE, UPPERCASE,
    DIGITS, HEXDIGITS, OCTDIGITS,
) = (
    'lettersdigits', 'letters', 'lowercase', 'uppercase',
    'digits', 'hexdigits', 'octdigits',
)

_char_class_members = {
    LETTERS_DIGITS: string.ascii_letters + string.digits,
    LETTERS: string.ascii_letters,
    LOWERCASE: string.ascii_lowercase,
    UPPERCASE: string.ascii_uppercase,
    DIGITS: string.digits,
    HEXDIGITS: string.digits + 'ABCDEF',
    OCTDIGITS: string.octdigits,
}


CharClass = collections.namedtuple('CharClass',
                                   ('allowed_chars', 'min_count'))


def named_char_class(char_class, min_count=0):
    """Return a predefined character class.

    The result of this function can be passed to :func:`generate_password` as
    one of the character classes to use in generating a password.

    :param char_class: Any of the character classes named in
                       :const:`CHARACTER_CLASSES`
    :param min_count: The minimum number of members of this class to appear in
                      a generated password
    """
    assert char_class in CHARACTER_CLASSES
    return CharClass(frozenset(_char_class_members[char_class]), min_count)


def special_char_class(allowed_chars, min_count=0):
    """Return a character class containing custom characters.

    The result of this function can be passed to :func:`generate_password` as
    one of the character classes to use in generating a password.

    :param allowed_chars: Iterable of the characters in the character class
    :param min_count: The minimum number of members of this class to appear in
                      a generated password
    """
    return CharClass(frozenset(allowed_chars), min_count)


def generate_password(length, char_classes):
    """Generate a random password.

    The password will be of the specified length, and comprised of characters
    from the specified character classes, which can be generated using the
    :func:`named_char_class` and :func:`special_char_class` functions. Where
    a minimum count is specified in the character class, at least that number
    of characters in the resulting password are guaranteed to be from that
    character class.

    :param length: The length of the password to generate, in characters
    :param char_classes: Iterable over classes of characters from which to
                         generate a password
    """
    char_buffer = io.StringIO()
    all_allowed_chars = set()

    # Add the minimum number of chars from each char class
    for char_class in char_classes:
        all_allowed_chars |= char_class.allowed_chars
        allowed_chars = tuple(char_class.allowed_chars)
        for i in range(char_class.min_count):
            char_buffer.write(random.choice(allowed_chars))

    # Fill up rest with random chars from provided classes
    combined_chars = tuple(all_allowed_chars)
    for i in range(max(0, length - char_buffer.tell())):
        char_buffer.write(random.choice(combined_chars))

    # Shuffle string
    selected_chars = char_buffer.getvalue()
    char_buffer.close()
    return ''.join(random.sample(selected_chars, length))


def generate_openstack_password():
    """Generate a random password suitable for a Keystone User."""
    return generate_password(32, [named_char_class(LOWERCASE, 1),
                                  named_char_class(UPPERCASE, 1),
                                  named_char_class(DIGITS, 1),
                                  special_char_class('!@#%^&*', 1)])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from ironicclient import exceptions as ic_exc

from heat.engine.clients.os import ironic as ic
from heat.tests import common
from heat.tests import utils


class IronicClientPluginTest(common.HeatTestCase):

    def test_create(self):
        context = utils.dummy_context()
        plugin = context.clients.client_plugin('ironic')
        client = plugin.client()
        self.assertEqual('http://server.test:5000/v3',
                         client.port.api.session.auth.endpoint)


class fake_resource(object):
    def __init__(self, id=None, name=None):
        self.uuid = id
        self.name = name


class PortGroupConstraintTest(common.HeatTestCase):
    def setUp(self):
        super(PortGroupConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_port_group_get = mock.Mock()
        self.ctx.clients.client_plugin(
            'ironic').client().portgroup.get = self.mock_port_group_get
        self.constraint = ic.PortGroupConstraint()

    def test_validate(self):
        self.mock_port_group_get.return_value = fake_resource(
            id='my_port_group')
        self.assertTrue(self.constraint.validate(
            'my_port_group', self.ctx))

    def test_validate_fail(self):
        self.mock_port_group_get.side_effect = ic_exc.NotFound()
        self.assertFalse(self.constraint.validate(
            "bad_port_group", self.ctx))


class NodeConstraintTest(common.HeatTestCase):
    def setUp(self):
        super(NodeConstraintTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.mock_node_get = mock.Mock()
        self.ctx.clients.client_plugin(
            'ironic').client().node.get = self.mock_node_get
        self.constraint = ic.NodeConstraint()

    def test_validate(self):
        self.mock_node_get.return_value = fake_resource(
            id='my_node')
        self.assertTrue(self.constraint.validate(
            'my_node', self.ctx))

    def test_validate_fail(self):
        self.mock_node_get.side_effect = ic_exc.NotFound()
        self.assertFalse(self.constraint.validate(
            "bad_node", self.ctx))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine import api as engine_api
from heat.engine import notification


def send(stack,
         adjustment=None,
         adjustment_type=None,
         capacity=None,
         groupname=None,
         message='error',
         suffix=None):
    """Send autoscaling notifications to the configured notification driver."""

    # see: https://wiki.openstack.org/wiki/SystemUsageData

    event_type = '%s.%s' % ('autoscaling', suffix)
    body = engine_api.format_notification_body(stack)
    body['adjustment_type'] = adjustment_type
    body['adjustment'] = adjustment
    body['capacity'] = capacity
    body['groupname'] = groupname
    body['message'] = message

    level = notification.get_default_level()
    if suffix == 'error':
        level = notification.ERROR

    notification.notify(stack.context, event_type, level, body)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from heat_integrationtests.functional import functional_base


class ReplaceDeprecatedResourceTest(functional_base.FunctionalTestsBase):
    template = '''
heat_template_version: "2013-05-23"
parameters:
  flavor:
    type: string
  image:
    type: string
  network:
    type: string

resources:
  config:
    type: OS::Heat::SoftwareConfig
    properties:
      config: xxxx

  server:
    type: OS::Nova::Server
    properties:
      image: {get_param: image}
      flavor: {get_param: flavor}
      networks: [{network: {get_param: network} }]
      user_data_format: SOFTWARE_CONFIG
  dep:
    type: OS::Heat::SoftwareDeployments
    properties:
        config: {get_resource: config}
        servers: {'0': {get_resource: server}}
        signal_transport: NO_SIGNAL
outputs:
  server:
    value: {get_resource: server}
'''

    deployment_group_snippet = '''
type: OS::Heat::SoftwareDeploymentGroup
properties:
  config: {get_resource: config}
  servers: {'0': {get_resource: server}}
  signal_transport: NO_SIGNAL
'''
    enable_cleanup = True

    def test_replace_software_deployments(self):
        parms = {'flavor': self.conf.minimal_instance_type,
                 'network': self.conf.fixed_network_name,
                 'image': self.conf.minimal_image_ref
                 }
        deployments_template = yaml.safe_load(self.template)
        stack_identifier = self.stack_create(
            parameters=parms,
            template=deployments_template,
            enable_cleanup=self.enable_cleanup)

        expected_resources = {'config': 'OS::Heat::SoftwareConfig',
                              'dep': 'OS::Heat::SoftwareDeployments',
                              'server': 'OS::Nova::Server'}
        self.assertEqual(expected_resources,
                         self.list_resources(stack_identifier))

        resource = self.client.resources.get(stack_identifier, 'dep')
        initial_phy_id = resource.physical_resource_id

        resources = deployments_template['resources']
        resources['dep'] = yaml.safe_load(self.deployment_group_snippet)
        self.update_stack(
            stack_identifier,
            deployments_template,
            parameters=parms)

        expected_new_resources = {'config': 'OS::Heat::SoftwareConfig',
                                  'dep': 'OS::Heat::SoftwareDeploymentGroup',
                                  'server': 'OS::Nova::Server'}
        self.assertEqual(expected_new_resources,
                         self.list_resources(stack_identifier))

        resource = self.client.resources.get(stack_identifier, 'dep')
        self.assertEqual(initial_phy_id, resource.physical_resource_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine.resources.openstack.senlin import res_base


class Receiver(res_base.BaseSenlinResource):
    """A resource that creates Senlin Receiver.

    Receiver is an abstract resource created at the senlin engine
    that can be used to hook the engine to some external event/alarm sources.
    """
    pass


def resource_mapping():
    return {
        'OS::Senlin::Receiver': Receiver
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine.resources.openstack.heat import none_resource
from heat.engine import support


class Bay(none_resource.NoneResource):
    """A resource that creates a Magnum Bay.

    This resource has been deprecated in favor of OS::Magnum::Cluster.
    """

    deprecation_msg = _('Please use OS::Magnum::Cluster instead.')
    support_status = support.SupportStatus(
        status=support.HIDDEN,
        message=deprecation_msg,
        version='11.0.0',
        previous_status=support.SupportStatus(
            status=support.DEPRECATED,
            message=deprecation_msg,
            version='9.0.0',
            previous_status=support.SupportStatus(
                status=support.SUPPORTED,
                version='6.0.0')
        )
    )


def resource_mapping():
    return {
        'OS::Magnum::Bay': Bay
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import weakref

from heat.common import context
from heat.common import exception
from heat.common.i18n import _
from heat.db import api as db_api
from heat.objects import raw_template_files

_d = weakref.WeakValueDictionary()


class ReadOnlyDict(dict):
    def __setitem__(self, key):
        raise ValueError("Attempted to write to internal TemplateFiles cache")


class TemplateFiles(collections.abc.Mapping):

    def __init__(self, files):
        self.files = None
        self.files_id = None
        if files is None:
            return
        if isinstance(files, TemplateFiles):
            self.files_id = files.files_id
            self.files = files.files
            return
        if isinstance(files, int):
            self.files_id = files
            if self.files_id in _d:
                self.files = _d[self.files_id]
            return
        if not isinstance(files, dict):
            raise ValueError(_('Expected dict, got %(cname)s for files, '
                               '(value is %(val)s)') %
                             {'cname': files.__class__,
                              'val': str(files)})
        # the dict has not been persisted as a raw_template_files DB obj
        # yet, so no self.files_id
        self.files = ReadOnlyDict(files)

    def __getitem__(self, key):
        self._refresh_if_needed()
        if self.files is None:
            raise KeyError
        return self.files[key]

    def __setitem__(self, key, value):
        self.update({key: value})

    def __len__(self):
        self._refresh_if_needed()
        if not self.files:
            return 0
        return len(self.files)

    def __contains__(self, key):
        self._refresh_if_needed()
        if not self.files:
            return False
        return key in self.files

    def __iter__(self):
        self._refresh_if_needed()
        if self.files is None:
            return iter(ReadOnlyDict({}))
        return iter(self.files)

    def _refresh_if_needed(self):
        # retrieve files from DB if needed
        if self.files_id is None:
            return
        if self.files_id in _d:
            self.files = _d[self.files_id]
            return
        self._refresh()

    def _refresh(self):
        ctxt = context.get_admin_context()
        rtf_obj = db_api.raw_template_files_get(ctxt, self.files_id)
        _files_dict = ReadOnlyDict(rtf_obj.files)
        self.files = _files_dict
        _d[self.files_id] = _files_dict

    def store(self, ctxt):
        if not self.files or self.files_id is not None:
            # Do not to persist an empty raw_template_files obj. If we
            # already have a not null self.files_id, the (immutable)
            # raw_templated_object has already been persisted so just
            # return the id.
            return self.files_id
        rtf_obj = raw_template_files.RawTemplateFiles.create(
            ctxt, {'files': self.files})
        self.files_id = rtf_obj.id
        _d[self.files_id] = self.files
        return self.files_id

    def update(self, files):
        # Sets up the next call to store() to create a new
        # raw_template_files DB obj. It seems like we *could* just
        # update the existing raw_template_files obj, but the problem
        # with that is other heat-engine processes' _d dictionaries
        # would have stale data for a given raw_template_files.id with
        # no way of knowing whether that data should be refreshed or
        # not. So, just avoid the potential for weird race conditions
        # and create another DB obj in the next store().
        if len(files) == 0:
            return
        if not isinstance(files, dict):
            raise ValueError(_('Expected dict, got %(cname)s for files, '
                               '(value is %(val)s)') %
                             {'cname': files.__class__,
                              'val': str(files)})

        self._refresh_if_needed()
        if self.files:
            new_files = self.files.copy()
            new_files.update(files)
        else:
            new_files = files
        self.files_id = None  # not persisted yet
        self.files = ReadOnlyDict(new_files)


def get_files_from_container(cnxt, files_container, files=None):

    if files is None:
        files = {}
    else:
        files = files.copy()

    swift_plugin = cnxt.clients.client_plugin('swift')

    if not swift_plugin:
        raise exception.ClientNotAvailable(client_name='swift')

    new_files = swift_plugin.get_files_from_container(files_container,
                                                      list(files.keys()))
    new_files.update(files)
    return new_files



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.resources.openstack.barbican import container
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils

stack_template_generic = '''
heat_template_version: 2015-10-15
description: Test template
resources:
  container:
    type: OS::Barbican::GenericContainer
    properties:
      name: mynewcontainer
      secrets:
        - name: secret1
          ref: ref1
        - name: secret2
          ref: ref2
'''

stack_template_certificate = '''
heat_template_version: 2015-10-15
description: Test template
resources:
  container:
    type: OS::Barbican::CertificateContainer
    properties:
      name: mynewcontainer
      certificate_ref: cref
      private_key_ref: pkref
      private_key_passphrase_ref: pkpref
      intermediates_ref: iref
'''

stack_template_rsa = '''
heat_template_version: 2015-10-15
description: Test template
resources:
  container:
    type: OS::Barbican::RSAContainer
    properties:
      name: mynewcontainer
      private_key_ref: pkref
      private_key_passphrase_ref: pkpref
      public_key_ref: pubref
'''


def template_by_name(name='OS::Barbican::GenericContainer'):
    mapping = {'OS::Barbican::GenericContainer': stack_template_generic,
               'OS::Barbican::CertificateContainer':
               stack_template_certificate,
               'OS::Barbican::RSAContainer': stack_template_rsa}
    return mapping[name]


class FakeContainer(object):

    def __init__(self, name):
        self.name = name

    def store(self):
        return self.name


class TestContainer(common.HeatTestCase):

    def setUp(self):
        super(TestContainer, self).setUp()

        self.patcher_client = mock.patch.object(
            container.GenericContainer, 'client')
        self.patcher_plugin = mock.patch.object(
            container.GenericContainer, 'client_plugin')
        mock_client = self.patcher_client.start()
        self.client = mock_client.return_value
        mock_plugin = self.patcher_plugin.start()
        self.client_plugin = mock_plugin.return_value
        self.stub_SecretConstraint_validate()

    def tearDown(self):
        super(TestContainer, self).tearDown()
        self.patcher_client.stop()
        self.patcher_plugin.stop()

    def _create_resource(self, name, snippet=None, stack=None,
                         tmpl_name='OS::Barbican::GenericContainer'):

        tmpl = template_format.parse(template_by_name(tmpl_name))
        if stack is None:
            self.stack = utils.parse_stack(tmpl)
        else:
            self.stack = stack
        resource_defns = self.stack.t.resource_definitions(self.stack)
        if snippet is None:
            snippet = resource_defns['container']
        res_class = container.resource_mapping()[tmpl_name]
        res = res_class(name, snippet, self.stack)
        res.check_create_complete = mock.Mock(return_value=True)
        create_generic_container = self.client_plugin.create_generic_container
        create_generic_container.return_value = FakeContainer('generic')
        self.client_plugin.create_certificate.return_value = FakeContainer(
            'certificate'
        )
        self.client_plugin.create_rsa.return_value = FakeContainer('rsa')
        scheduler.TaskRunner(res.create)()
        return res

    def test_create_generic(self):
        res = self._create_resource('foo')
        expected_state = (res.CREATE, res.COMPLETE)
        self.assertEqual(expected_state, res.state)
        args = self.client_plugin.create_generic_container.call_args[1]
        self.assertEqual('mynewcontainer', args['name'])
        self.assertEqual({'secret1': 'ref1', 'secret2': 'ref2'},
                         args['secret_refs'])
        self.assertEqual(sorted(['ref1', 'ref2']), sorted(res.get_refs()))

    def test_create_certificate(self):
        res = self._create_resource(
            'foo', tmpl_name='OS::Barbican::CertificateContainer')
        expected_state = (res.CREATE, res.COMPLETE)
        self.assertEqual(expected_state, res.state)
        args = self.client_plugin.create_certificate.call_args[1]
        self.assertEqual('mynewcontainer', args['name'])
        self.assertEqual('cref', args['certificate_ref'])
        self.assertEqual('pkref', args['private_key_ref'])
        self.assertEqual('pkpref', args['private_key_passphrase_ref'])
        self.assertEqual('iref', args['intermediates_ref'])
        self.assertEqual(sorted(['pkref', 'pkpref', 'iref', 'cref']),
                         sorted(res.get_refs()))

    def test_create_rsa(self):
        res = self._create_resource(
            'foo', tmpl_name='OS::Barbican::RSAContainer')
        expected_state = (res.CREATE, res.COMPLETE)
        self.assertEqual(expected_state, res.state)
        args = self.client_plugin.create_rsa.call_args[1]
        self.assertEqual('mynewcontainer', args['name'])
        self.assertEqual('pkref', args['private_key_ref'])
        self.assertEqual('pubref', args['public_key_ref'])
        self.assertEqual('pkpref', args['private_key_passphrase_ref'])
        self.assertEqual(sorted(['pkref', 'pubref', 'pkpref']),
                         sorted(res.get_refs()))

    def test_create_failed_on_validation(self):
        tmpl = template_format.parse(template_by_name())
        stack = utils.parse_stack(tmpl)
        props = tmpl['resources']['container']['properties']
        props['secrets'].append({'name': 'secret3', 'ref': 'ref1'})
        defn = rsrc_defn.ResourceDefinition(
            'failed_container', 'OS::Barbican::GenericContainer', props)
        res = container.GenericContainer('foo', defn, stack)
        self.assertRaisesRegex(exception.StackValidationFailed,
                               'Duplicate refs are not allowed',
                               res.validate)

    def test_attributes(self):
        mock_container = mock.Mock()
        mock_container.status = 'test-status'
        mock_container.container_ref = 'test-container-ref'
        mock_container.secret_refs = {'name': 'ref'}
        mock_container.consumers = [{'name': 'name1', 'ref': 'ref1'}]
        res = self._create_resource('foo')
        self.client.containers.get.return_value = mock_container
        self.assertEqual('test-status', res.FnGetAtt('status'))
        self.assertEqual('test-container-ref', res.FnGetAtt('container_ref'))
        self.assertEqual({'name': 'ref'}, res.FnGetAtt('secret_refs'))
        self.assertEqual([{'name': 'name1', 'ref': 'ref1'}],
                         res.FnGetAtt('consumers'))

    def test_check_create_complete(self):
        tmpl = template_format.parse(template_by_name())
        stack = utils.parse_stack(tmpl)
        resource_defns = stack.t.resource_definitions(stack)
        res_template = resource_defns['container']
        res = container.GenericContainer('foo', res_template, stack)
        mock_active = mock.Mock(status='ACTIVE')
        self.client.containers.get.return_value = mock_active
        self.assertTrue(res.check_create_complete('foo'))
        mock_not_active = mock.Mock(status='PENDING')
        self.client.containers.get.return_value = mock_not_active
        self.assertFalse(res.check_create_complete('foo'))
        mock_not_active = mock.Mock(status='ERROR', error_reason='foo',
                                    error_status_code=500)
        self.client.containers.get.return_value = mock_not_active
        exc = self.assertRaises(exception.ResourceInError,
                                res.check_create_complete, 'foo')
        self.assertIn('foo', str(exc))
        self.assertIn('500', str(exc))
